@@ -7,7 +7,6 @@
 #include "types.h"
 #include "memory.h"
 #include "cpt.h"
-#include "shared.h"
 
 int cpt_ctx_init (hashcat_ctx_t *hashcat_ctx)
 {
@@ -19,7 +18,7 @@ int cpt_ctx_init (hashcat_ctx_t *hashcat_ctx)
   if (user_options->example_hashes == true) return 0;
   if (user_options->keyspace       == true) return 0;
   if (user_options->left           == true) return 0;
-  if (user_options->opencl_info    == true) return 0;
+  if (user_options->backend_info   == true) return 0;
   if (user_options->show           == true) return 0;
   if (user_options->usage          == true) return 0;
   if (user_options->version        == true) return 0;
