@@ -11,8 +11,7 @@
 #include "logfile.h"
 #include "convert.h"
 #include "filehandling.h"
-#include "interface.h"
-#include "opencl.h"
+#include "backend.h"
 #include "shared.h"
 #include "ext_lzma.h"
 #include "mpsp.h"
@@ -118,23 +117,16 @@ static int mp_css_append_salt (hashcat_ctx_t *hashcat_ctx, salt_t *salt_buf)
   u32  salt_len     =        salt_buf->salt_len;
   u8  *salt_buf_ptr = (u8 *) salt_buf->salt_buf;
 
-  u32 css_cnt_salt = mask_ctx->css_cnt + salt_len;
-
-  cs_t *css_buf_salt = (cs_t *) hccalloc (css_cnt_salt, sizeof (cs_t));
-
-  memcpy (css_buf_salt, mask_ctx->css_buf, mask_ctx->css_cnt * sizeof (cs_t));
+  if ((mask_ctx->css_cnt + salt_len) > 256) return -1;
 
   for (u32 i = 0, j = mask_ctx->css_cnt; i < salt_len; i++, j++)
   {
-    css_buf_salt[j].cs_buf[0] = salt_buf_ptr[i];
-    css_buf_salt[j].cs_len    = 1;
+    mask_ctx->css_buf[j].cs_buf[0] = salt_buf_ptr[i];
+    mask_ctx->css_buf[j].cs_len    = 1;
+
+    mask_ctx->css_cnt++;
   }
 
-  hcfree (mask_ctx->css_buf);
-
-  mask_ctx->css_buf = css_buf_salt;
-  mask_ctx->css_cnt = css_cnt_salt;
-
   return 0;
 }
 
@@ -144,6 +136,8 @@ static int mp_css_utf16le_expand (hashcat_ctx_t *hashcat_ctx)
 
   u32 css_cnt_utf16le = mask_ctx->css_cnt * 2;
 
+  if (css_cnt_utf16le > 256) return -1;
+
   cs_t *css_buf_utf16le = (cs_t *) hccalloc (css_cnt_utf16le, sizeof (cs_t));
 
   for (u32 i = 0, j = 0; i < mask_ctx->css_cnt; i += 1, j += 2)
@@ -154,11 +148,12 @@ static int mp_css_utf16le_expand (hashcat_ctx_t *hashcat_ctx)
     css_buf_utf16le[j + 1].cs_len    = 1;
   }
 
-  hcfree (mask_ctx->css_buf);
+  memcpy (mask_ctx->css_buf, css_buf_utf16le, css_cnt_utf16le * sizeof (cs_t));
 
-  mask_ctx->css_buf = css_buf_utf16le;
   mask_ctx->css_cnt = css_cnt_utf16le;
 
+  hcfree (css_buf_utf16le);
+
   return 0;
 }
 
@@ -168,6 +163,8 @@ static int mp_css_utf16be_expand (hashcat_ctx_t *hashcat_ctx)
 
   u32 css_cnt_utf16be = mask_ctx->css_cnt * 2;
 
+  if (css_cnt_utf16be > 256) return -1;
+
   cs_t *css_buf_utf16be = (cs_t *) hccalloc (css_cnt_utf16be, sizeof (cs_t));
 
   for (u32 i = 0, j = 0; i < mask_ctx->css_cnt; i += 1, j += 2)
@@ -178,11 +175,12 @@ static int mp_css_utf16be_expand (hashcat_ctx_t *hashcat_ctx)
     memcpy (&css_buf_utf16be[j + 1], &mask_ctx->css_buf[i], sizeof (cs_t));
   }
 
-  hcfree (mask_ctx->css_buf);
+  memcpy (mask_ctx->css_buf, css_buf_utf16be, css_cnt_utf16be * sizeof (cs_t));
 
-  mask_ctx->css_buf = css_buf_utf16be;
   mask_ctx->css_cnt = css_cnt_utf16be;
 
+  hcfree (css_buf_utf16be);
+
   return 0;
 }
 
@@ -219,6 +217,13 @@ static int mp_add_cs_buf (hashcat_ctx_t *hashcat_ctx, const u32 *in_buf, size_t
 {
   const hashconfig_t *hashconfig = hashcat_ctx->hashconfig;
 
+  if (css_cnt == 256)
+  {
+    event_log_error (hashcat_ctx, "Invalid mask length.");
+
+    return -1;
+  }
+
   cs_t *cs = &css[css_cnt];
 
   size_t css_uniq_sz = CHARSIZ * sizeof (u32);
@@ -256,7 +261,7 @@ static int mp_add_cs_buf (hashcat_ctx_t *hashcat_ctx, const u32 *in_buf, size_t
 
 static int mp_expand (hashcat_ctx_t *hashcat_ctx, const char *in_buf, size_t in_len, cs_t *mp_sys, cs_t *mp_usr, u32 mp_usr_offset, int interpret)
 {
-  const user_options_t *user_options = hashcat_ctx->user_options;
+  const hashconfig_t *hashconfig = hashcat_ctx->hashconfig;
 
   size_t in_pos;
 
@@ -319,7 +324,7 @@ static int mp_expand (hashcat_ctx_t *hashcat_ctx, const char *in_buf, size_t in_
     }
     else
     {
-      if (user_options->hex_charset == true)
+      if (hashconfig->opts_type & OPTS_TYPE_PT_HEX)
       {
         in_pos++;
 
@@ -364,10 +369,12 @@ static int mp_expand (hashcat_ctx_t *hashcat_ctx, const char *in_buf, size_t in_
 
 static int mp_gen_css (hashcat_ctx_t *hashcat_ctx, char *mask_buf, size_t mask_len, cs_t *mp_sys, cs_t *mp_usr, cs_t *css_buf, u32 *css_cnt)
 {
-  const user_options_t *user_options = hashcat_ctx->user_options;
+  const hashconfig_t *hashconfig = hashcat_ctx->hashconfig;
 
-  u32 mask_pos;
-  u32 css_pos;
+  memset (css_buf, 0, 256 * sizeof (cs_t));
+
+  size_t mask_pos;
+  size_t css_pos;
 
   for (mask_pos = 0, css_pos = 0; mask_pos < mask_len; mask_pos++, css_pos++)
   {
@@ -430,7 +437,7 @@ static int mp_gen_css (hashcat_ctx_t *hashcat_ctx, char *mask_buf, size_t mask_l
     }
     else
     {
-      if (user_options->hex_charset == true)
+      if (hashconfig->opts_type & OPTS_TYPE_PT_HEX)
       {
         mask_pos++;
 
@@ -488,7 +495,7 @@ static int mp_gen_css (hashcat_ctx_t *hashcat_ctx, char *mask_buf, size_t mask_l
 
 static int mp_get_truncated_mask (hashcat_ctx_t *hashcat_ctx, const char *mask_buf, const size_t mask_len, const u32 len, char *new_mask_buf)
 {
-  const user_options_t *user_options = hashcat_ctx->user_options;
+  const hashconfig_t *hashconfig = hashcat_ctx->hashconfig;
 
   u32 mask_pos;
 
@@ -512,7 +519,7 @@ static int mp_get_truncated_mask (hashcat_ctx_t *hashcat_ctx, const char *mask_b
     }
     else
     {
-      if (user_options->hex_charset == true)
+      if (hashconfig->opts_type & OPTS_TYPE_PT_HEX)
       {
         mask_pos++;
 
@@ -583,9 +590,9 @@ static void mp_setup_sys (cs_t *mp_sys)
 
 static int mp_setup_usr (hashcat_ctx_t *hashcat_ctx, cs_t *mp_sys, cs_t *mp_usr, const char *buf, const u32 userindex)
 {
-  FILE *fp = fopen (buf, "rb");
+  HCFILE fp;
 
-  if (fp == NULL) // feof() in case if file is empty
+  if (hc_fopen (&fp, buf, "rb") == false)
   {
     const int rc = mp_expand (hashcat_ctx, buf, strlen (buf), mp_sys, mp_usr, userindex, 1);
 
@@ -595,18 +602,18 @@ static int mp_setup_usr (hashcat_ctx_t *hashcat_ctx, cs_t *mp_sys, cs_t *mp_usr,
   {
     char mp_file[1024];
 
-    const size_t nread = hc_fread (mp_file, 1, sizeof (mp_file) - 1, fp);
+    const size_t nread = hc_fread (mp_file, 1, sizeof (mp_file) - 1, &fp);
 
-    if (!feof (fp))
+    if (!hc_feof (&fp))
     {
       event_log_error (hashcat_ctx, "%s: Custom charset file is too large.", buf);
 
-      fclose (fp);
+      hc_fclose (&fp);
 
       return -1;
     }
 
-    fclose (fp);
+    hc_fclose (&fp);
 
     if (nread == 0)
     {
@@ -711,9 +718,9 @@ static int sp_setup_tbl (hashcat_ctx_t *hashcat_ctx)
     return -1;
   }
 
-  FILE *fd = fopen (hcstat, "rb");
+  HCFILE fp;
 
-  if (fd == NULL)
+  if (hc_fopen (&fp, hcstat, "rb") == false)
   {
     event_log_error (hashcat_ctx, "%s: %s", hcstat, strerror (errno));
 
@@ -722,20 +729,20 @@ static int sp_setup_tbl (hashcat_ctx_t *hashcat_ctx)
 
   u8 *inbuf = (u8 *) hcmalloc (s.st_size);
 
-  SizeT inlen = (SizeT) hc_fread (inbuf, 1, s.st_size, fd);
+  SizeT inlen = (SizeT) hc_fread (inbuf, 1, s.st_size, &fp);
 
   if (inlen != (SizeT) s.st_size)
   {
     event_log_error (hashcat_ctx, "%s: Could not read data.", hcstat);
 
-    fclose (fd);
+    hc_fclose (&fp);
 
     hcfree (inbuf);
 
     return -1;
   }
 
-  fclose (fd);
+  hc_fclose (&fp);
 
   u8 *outbuf = (u8 *) hcmalloc (SP_FILESZ);
 
@@ -1061,7 +1068,7 @@ static int mask_append (hashcat_ctx_t *hashcat_ctx, const char *mask, const char
 
   if (user_options->increment == true)
   {
-    const u32 mask_length = mp_get_length (mask);
+    const u32 mask_length = mp_get_length (mask, hashconfig->opts_type);
 
     u32 increment_min = user_options->increment_min;
     u32 increment_max = user_options->increment_max;
@@ -1090,9 +1097,7 @@ static int mask_append (hashcat_ctx_t *hashcat_ctx, const char *mask, const char
         mask_truncated_next += snprintf (mask_truncated, 256, "%s,", prepend);
       }
 
-      const int rc_truncated_mask = mp_get_truncated_mask (hashcat_ctx, mask, strlen (mask), increment_len, mask_truncated_next);
-
-      if (rc_truncated_mask == -1)
+      if (mp_get_truncated_mask (hashcat_ctx, mask, strlen (mask), increment_len, mask_truncated_next) == -1)
       {
         hcfree (mask_truncated);
 
@@ -1112,7 +1117,7 @@ static int mask_append (hashcat_ctx_t *hashcat_ctx, const char *mask, const char
     {
       // this happens with maskfiles only
 
-      char *prepend_mask;
+      char *prepend_mask = NULL;
 
       hc_asprintf (&prepend_mask, "%s,%s", prepend, mask);
 
@@ -1124,26 +1129,41 @@ static int mask_append (hashcat_ctx_t *hashcat_ctx, const char *mask, const char
     }
     else
     {
-      const int rc = mask_append_final (hashcat_ctx, mask);
-
-      if (rc == -1) return -1;
+      if (mask_append_final (hashcat_ctx, mask) == -1) return -1;
     }
   }
 
   return 0;
 }
 
-u32 mp_get_length (const char *mask)
+u32 mp_get_length (const char *mask, const u32 opts_type)
 {
+  bool ignore_next = false;
+
   u32 len = 0;
 
   const size_t mask_len = strlen (mask);
 
   for (size_t i = 0; i < mask_len; i++)
   {
-    if (mask[i] == '?') i++;
+    if (ignore_next == true)
+    {
+      ignore_next = false;
+    }
+    else
+    {
+      if (mask[i] == '?')
+      {
+        ignore_next = true;
+      }
 
-    len++;
+      if (opts_type & OPTS_TYPE_PT_HEX)
+      {
+        ignore_next = true;
+      }
+
+      len++;
+    }
   }
 
   return len;
@@ -1188,6 +1208,7 @@ int mask_ctx_update_loop (hashcat_ctx_t *hashcat_ctx)
   user_options_extra_t *user_options_extra = hashcat_ctx->user_options_extra;
   user_options_t       *user_options       = hashcat_ctx->user_options;
 
+
   if (user_options_extra->attack_kern == ATTACK_KERN_COMBI)
   {
     if (user_options->attack_mode == ATTACK_MODE_COMBI)
@@ -1200,15 +1221,9 @@ int mask_ctx_update_loop (hashcat_ctx_t *hashcat_ctx)
       {
         mask_ctx->mask = mask_ctx->masks[mask_ctx->masks_pos];
 
-        const int rc_mask_file = mask_ctx_parse_maskfile (hashcat_ctx);
+        if (mask_ctx_parse_maskfile (hashcat_ctx) == -1) return -1;
 
-        if (rc_mask_file == -1) return -1;
-
-        mask_ctx->css_buf = (cs_t *) hccalloc (256, sizeof (cs_t));
-
-        const int rc_gen_css = mp_gen_css (hashcat_ctx, mask_ctx->mask, strlen (mask_ctx->mask), mask_ctx->mp_sys, mask_ctx->mp_usr, mask_ctx->css_buf, &mask_ctx->css_cnt);
-
-        if (rc_gen_css == -1) return -1;
+        if (mp_gen_css (hashcat_ctx, mask_ctx->mask, strlen (mask_ctx->mask), mask_ctx->mp_sys, mask_ctx->mp_usr, mask_ctx->css_buf, &mask_ctx->css_cnt) == -1) return -1;
 
         u32 uniq_tbls[SP_PW_MAX][CHARSIZ] = { { 0 } };
 
@@ -1216,32 +1231,22 @@ int mask_ctx_update_loop (hashcat_ctx_t *hashcat_ctx)
 
         sp_tbl_to_css (mask_ctx->root_table_buf, mask_ctx->markov_table_buf, mask_ctx->root_css_buf, mask_ctx->markov_css_buf, user_options->markov_threshold, uniq_tbls);
 
-        const int rc_get_sum = sp_get_sum (0, mask_ctx->css_cnt, mask_ctx->root_css_buf, &mask_ctx->bfs_cnt);
-
-        if (rc_get_sum == -1)
+        if (sp_get_sum (0, mask_ctx->css_cnt, mask_ctx->root_css_buf, &mask_ctx->bfs_cnt) == -1)
         {
           event_log_error (hashcat_ctx, "Integer overflow detected in keyspace of mask: %s", mask_ctx->mask);
 
           return -1;
         }
 
-        const int rc_update_mp = opencl_session_update_mp (hashcat_ctx);
-
-        if (rc_update_mp == -1) return -1;
+        if (backend_session_update_mp (hashcat_ctx) == -1) return -1;
       }
       else
       {
         mask_ctx->mask = mask_ctx->masks[mask_ctx->masks_pos];
 
-        const int rc_mask_file = mask_ctx_parse_maskfile (hashcat_ctx);
+        if (mask_ctx_parse_maskfile (hashcat_ctx) == -1) return -1;
 
-        if (rc_mask_file == -1) return -1;
-
-        mask_ctx->css_buf = (cs_t *) hccalloc (256, sizeof (cs_t));
-
-        const int rc_gen_css = mp_gen_css (hashcat_ctx, mask_ctx->mask, strlen (mask_ctx->mask), mask_ctx->mp_sys, mask_ctx->mp_usr, mask_ctx->css_buf, &mask_ctx->css_cnt);
-
-        if (rc_gen_css == -1) return -1;
+        if (mp_gen_css (hashcat_ctx, mask_ctx->mask, strlen (mask_ctx->mask), mask_ctx->mp_sys, mask_ctx->mp_usr, mask_ctx->css_buf, &mask_ctx->css_cnt) == -1) return -1;
 
         u32 uniq_tbls[SP_PW_MAX][CHARSIZ] = { { 0 } };
 
@@ -1249,40 +1254,38 @@ int mask_ctx_update_loop (hashcat_ctx_t *hashcat_ctx)
 
         sp_tbl_to_css (mask_ctx->root_table_buf, mask_ctx->markov_table_buf, mask_ctx->root_css_buf, mask_ctx->markov_css_buf, user_options->markov_threshold, uniq_tbls);
 
-        const int rc_get_sum = sp_get_sum (0, mask_ctx->css_cnt, mask_ctx->root_css_buf, &combinator_ctx->combs_cnt);
-
-        if (rc_get_sum == -1)
+        if (sp_get_sum (0, mask_ctx->css_cnt, mask_ctx->root_css_buf, &combinator_ctx->combs_cnt) == -1)
         {
           event_log_error (hashcat_ctx, "Integer overflow detected in keyspace of mask: %s", mask_ctx->mask);
 
           return -1;
         }
 
-        const int rc_update_mp = opencl_session_update_mp (hashcat_ctx);
+        // do not allow modifier count > 32 bit
+        // https://github.com/hashcat/hashcat/issues/2482
 
-        if (rc_update_mp == -1) return -1;
+        if (combinator_ctx->combs_cnt > 0xffffffff)
+        {
+          event_log_error (hashcat_ctx, "Integer overflow detected in keyspace of mask: %s", mask_ctx->mask);
+
+          return -1;
+        }
+
+        if (backend_session_update_mp (hashcat_ctx) == -1) return -1;
       }
     }
 
-    const int rc_update_combinator = opencl_session_update_combinator (hashcat_ctx);
-
-    if (rc_update_combinator == -1) return -1;
+    if (backend_session_update_combinator (hashcat_ctx) == -1) return -1;
   }
   else if (user_options_extra->attack_kern == ATTACK_KERN_BF)
   {
     mask_ctx->mask = mask_ctx->masks[mask_ctx->masks_pos];
 
-    const int rc_mask_file = mask_ctx_parse_maskfile (hashcat_ctx);
-
-    if (rc_mask_file == -1) return -1;
+    if (mask_ctx_parse_maskfile (hashcat_ctx) == -1) return -1;
 
     if (user_options->attack_mode == ATTACK_MODE_BF) // always true
     {
-      mask_ctx->css_buf = (cs_t *) hccalloc (256, sizeof (cs_t));
-
-      const int rc_gen_css = mp_gen_css (hashcat_ctx, mask_ctx->mask, strlen (mask_ctx->mask), mask_ctx->mp_sys, mask_ctx->mp_usr, mask_ctx->css_buf, &mask_ctx->css_cnt);
-
-      if (rc_gen_css == -1) return -1;
+      if (mp_gen_css (hashcat_ctx, mask_ctx->mask, strlen (mask_ctx->mask), mask_ctx->mp_sys, mask_ctx->mp_usr, mask_ctx->css_buf, &mask_ctx->css_cnt) == -1) return -1;
 
       // special case for benchmark
 
@@ -1291,7 +1294,7 @@ int mask_ctx_update_loop (hashcat_ctx_t *hashcat_ctx)
 
       if (user_options->benchmark == true)
       {
-        pw_min = mp_get_length (mask_ctx->mask);
+        pw_min = mp_get_length (mask_ctx->mask, hashconfig->opts_type);
         pw_max = pw_min;
       }
 
@@ -1324,17 +1327,16 @@ int mask_ctx_update_loop (hashcat_ctx_t *hashcat_ctx)
         return -1;
       }
 
-      if (hashconfig->opts_type & OPTS_TYPE_PT_UTF16LE)
+      if (user_options->slow_candidates == false)
       {
-        const int rc = mp_css_utf16le_expand (hashcat_ctx);
-
-        if (rc == -1) return -1;
-      }
-      else if (hashconfig->opts_type & OPTS_TYPE_PT_UTF16BE)
-      {
-        const int rc = mp_css_utf16be_expand (hashcat_ctx);
-
-        if (rc == -1) return -1;
+        if (hashconfig->opts_type & OPTS_TYPE_PT_UTF16LE)
+        {
+          if (mp_css_utf16le_expand (hashcat_ctx) == -1) return -1;
+        }
+        else if (hashconfig->opts_type & OPTS_TYPE_PT_UTF16BE)
+        {
+          if (mp_css_utf16be_expand (hashcat_ctx) == -1) return -1;
+        }
       }
 
       u32 css_cnt_orig = mask_ctx->css_cnt;
@@ -1343,9 +1345,7 @@ int mask_ctx_update_loop (hashcat_ctx_t *hashcat_ctx)
       {
         if (hashconfig->opti_type & OPTI_TYPE_APPENDED_SALT)
         {
-          const int rc = mp_css_append_salt (hashcat_ctx, &hashes->salts_buf[0]);
-
-          if (rc == -1) return -1;
+          if (mp_css_append_salt (hashcat_ctx, &hashes->salts_buf[0]) == -1) return -1;
         }
       }
 
@@ -1355,9 +1355,7 @@ int mask_ctx_update_loop (hashcat_ctx_t *hashcat_ctx)
 
       sp_tbl_to_css (mask_ctx->root_table_buf, mask_ctx->markov_table_buf, mask_ctx->root_css_buf, mask_ctx->markov_css_buf, user_options->markov_threshold, uniq_tbls);
 
-      const int rc_get_sum1 = sp_get_sum (0, mask_ctx->css_cnt, mask_ctx->root_css_buf, &status_ctx->words_cnt);
-
-      if (rc_get_sum1 == -1)
+      if (sp_get_sum (0, mask_ctx->css_cnt, mask_ctx->root_css_buf, &status_ctx->words_cnt) == -1)
       {
         event_log_error (hashcat_ctx, "Integer overflow detected in keyspace of mask: %s", mask_ctx->mask);
 
@@ -1370,18 +1368,14 @@ int mask_ctx_update_loop (hashcat_ctx_t *hashcat_ctx)
 
       mp_css_split_cnt (hashcat_ctx, css_cnt_orig, css_cnt_lr);
 
-      const int rc_get_sum2 = sp_get_sum (0, css_cnt_lr[1], mask_ctx->root_css_buf, &mask_ctx->bfs_cnt);
-
-      if (rc_get_sum2 == -1)
+      if (sp_get_sum (0, css_cnt_lr[1], mask_ctx->root_css_buf, &mask_ctx->bfs_cnt) == -1)
       {
         event_log_error (hashcat_ctx, "Integer overflow detected in keyspace of mask: %s", mask_ctx->mask);
 
         return -1;
       }
 
-      const int rc_update_mp_rl = opencl_session_update_mp_rl (hashcat_ctx, css_cnt_lr[0], css_cnt_lr[1]);
-
-      if (rc_update_mp_rl == -1) return -1;
+      if (backend_session_update_mp_rl (hashcat_ctx, css_cnt_lr[0], css_cnt_lr[1]) == -1) return -1;
     }
   }
 
@@ -1390,15 +1384,16 @@ int mask_ctx_update_loop (hashcat_ctx_t *hashcat_ctx)
 
 int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
 {
-  mask_ctx_t           *mask_ctx            = hashcat_ctx->mask_ctx;
-  user_options_extra_t *user_options_extra  = hashcat_ctx->user_options_extra;
-  user_options_t       *user_options        = hashcat_ctx->user_options;
+  const hashconfig_t         *hashconfig          = hashcat_ctx->hashconfig;
+  const user_options_extra_t *user_options_extra  = hashcat_ctx->user_options_extra;
+  const user_options_t       *user_options        = hashcat_ctx->user_options;
+  mask_ctx_t                 *mask_ctx            = hashcat_ctx->mask_ctx;
 
   mask_ctx->enabled = false;
 
   if (user_options->example_hashes == true) return 0;
   if (user_options->left           == true) return 0;
-  if (user_options->opencl_info    == true) return 0;
+  if (user_options->backend_info   == true) return 0;
   if (user_options->show           == true) return 0;
   if (user_options->usage          == true) return 0;
   if (user_options->version        == true) return 0;
@@ -1408,19 +1403,20 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
 
   mask_ctx->enabled = true;
 
+  mask_ctx->mp_sys  = (cs_t *) hccalloc (8, sizeof (cs_t));
+  mask_ctx->mp_usr  = (cs_t *) hccalloc (4, sizeof (cs_t));
+
+  mask_ctx->css_buf = (cs_t *) hccalloc (256, sizeof (cs_t));
+  mask_ctx->css_cnt = 0;
+
   mask_ctx->root_table_buf   = (hcstat_table_t *) hccalloc (SP_ROOT_CNT,   sizeof (hcstat_table_t));
   mask_ctx->markov_table_buf = (hcstat_table_t *) hccalloc (SP_MARKOV_CNT, sizeof (hcstat_table_t));
 
-  const int rc_setup_tbl = sp_setup_tbl (hashcat_ctx);
-
-  if (rc_setup_tbl == -1) return -1;
+  if (sp_setup_tbl (hashcat_ctx) == -1) return -1;
 
   mask_ctx->root_css_buf   = (cs_t *) hccalloc (SP_PW_MAX,           sizeof (cs_t));
   mask_ctx->markov_css_buf = (cs_t *) hccalloc (SP_PW_MAX * CHARSIZ, sizeof (cs_t));
 
-  mask_ctx->css_cnt = 0;
-  mask_ctx->css_buf = NULL;
-
   mask_ctx->mask_from_file = false;
 
   mask_ctx->masks     = NULL;
@@ -1431,10 +1427,10 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
 
   mp_setup_sys (mask_ctx->mp_sys);
 
-  if (user_options->custom_charset_1) { const int rc = mp_setup_usr (hashcat_ctx, mask_ctx->mp_sys, mask_ctx->mp_usr, user_options->custom_charset_1, 0); if (rc == -1) return -1; }
-  if (user_options->custom_charset_2) { const int rc = mp_setup_usr (hashcat_ctx, mask_ctx->mp_sys, mask_ctx->mp_usr, user_options->custom_charset_2, 1); if (rc == -1) return -1; }
-  if (user_options->custom_charset_3) { const int rc = mp_setup_usr (hashcat_ctx, mask_ctx->mp_sys, mask_ctx->mp_usr, user_options->custom_charset_3, 2); if (rc == -1) return -1; }
-  if (user_options->custom_charset_4) { const int rc = mp_setup_usr (hashcat_ctx, mask_ctx->mp_sys, mask_ctx->mp_usr, user_options->custom_charset_4, 3); if (rc == -1) return -1; }
+  if (user_options->custom_charset_1) { if (mp_setup_usr (hashcat_ctx, mask_ctx->mp_sys, mask_ctx->mp_usr, user_options->custom_charset_1, 0) == -1) return -1; }
+  if (user_options->custom_charset_2) { if (mp_setup_usr (hashcat_ctx, mask_ctx->mp_sys, mask_ctx->mp_usr, user_options->custom_charset_2, 1) == -1) return -1; }
+  if (user_options->custom_charset_3) { if (mp_setup_usr (hashcat_ctx, mask_ctx->mp_sys, mask_ctx->mp_usr, user_options->custom_charset_3, 2) == -1) return -1; }
+  if (user_options->custom_charset_4) { if (mp_setup_usr (hashcat_ctx, mask_ctx->mp_sys, mask_ctx->mp_usr, user_options->custom_charset_4, 3) == -1) return -1; }
 
   if (user_options->attack_mode == ATTACK_MODE_BF)
   {
@@ -1446,9 +1442,7 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
 
         if (hc_path_exist (arg) == false)
         {
-          const int rc = mask_append (hashcat_ctx, arg, NULL);
-
-          if (rc == -1) return -1;
+          if (mask_append (hashcat_ctx, arg, NULL) == -1) return -1;
         }
         else
         {
@@ -1460,9 +1454,9 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
 
             if (hc_path_is_file (arg) == true)
             {
-              FILE *mask_fp = fopen (arg, "r");
+              HCFILE mask_fp;
 
-              if (mask_fp == NULL)
+              if (hc_fopen (&mask_fp, arg, "r") == false)
               {
                 event_log_error (hashcat_ctx, "%s: %s", arg, strerror (errno));
 
@@ -1471,9 +1465,9 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
 
               char *line_buf = (char *) hcmalloc (HCBUFSIZ_LARGE);
 
-              while (!feof (mask_fp))
+              while (!hc_feof (&mask_fp))
               {
-                const size_t line_len = fgetl (mask_fp, line_buf);
+                const size_t line_len = fgetl (&mask_fp, line_buf, HCBUFSIZ_LARGE);
 
                 if (line_len == 0) continue;
 
@@ -1492,11 +1486,9 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
                   mask_buf[-1] = 0;
                 }
 
-                const int rc = mask_append (hashcat_ctx, mask_buf, prepend_buf);
-
-                if (rc == -1)
+                if (mask_append (hashcat_ctx, mask_buf, prepend_buf) == -1)
                 {
-                  fclose (mask_fp);
+                  hc_fclose (&mask_fp);
 
                   return -1;
                 }
@@ -1504,7 +1496,7 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
 
               hcfree (line_buf);
 
-              fclose (mask_fp);
+              hc_fclose (&mask_fp);
             }
             else
             {
@@ -1519,18 +1511,14 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
       {
         const char *mask = DEF_MASK;
 
-        const int rc = mask_append (hashcat_ctx, mask, NULL);
-
-        if (rc == -1) return -1;
+        if (mask_append (hashcat_ctx, mask, NULL) == -1) return -1;
       }
     }
     else
     {
-      const char *mask = hashconfig_benchmark_mask (hashcat_ctx);
+      const char *mask = hashconfig->benchmark_mask;
 
-      const int rc = mask_append (hashcat_ctx, mask, NULL);
-
-      if (rc == -1) return -1;
+      if (mask_append (hashcat_ctx, mask, NULL) == -1) return -1;
     }
   }
   else if (user_options->attack_mode == ATTACK_MODE_HYBRID1)
@@ -1543,9 +1531,7 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
 
     if (hc_path_exist (arg) == false)
     {
-      const int rc = mask_append (hashcat_ctx, arg, NULL);
-
-      if (rc == -1) return -1;
+      if (mask_append (hashcat_ctx, arg, NULL) == -1) return -1;
     }
     else
     {
@@ -1553,9 +1539,9 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
       {
         mask_ctx->mask_from_file = true;
 
-        FILE *mask_fp = fopen (arg, "r");
+        HCFILE mask_fp;
 
-        if (mask_fp == NULL)
+        if (hc_fopen (&mask_fp, arg, "r") == false)
         {
           event_log_error (hashcat_ctx, "%s: %s", arg, strerror (errno));
 
@@ -1564,9 +1550,9 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
 
         char *line_buf = (char *) hcmalloc (HCBUFSIZ_LARGE);
 
-        while (!feof (mask_fp))
+        while (!hc_feof (&mask_fp))
         {
-          const size_t line_len = fgetl (mask_fp, line_buf);
+          const size_t line_len = fgetl (&mask_fp, line_buf, HCBUFSIZ_LARGE);
 
           if (line_len == 0) continue;
 
@@ -1585,11 +1571,9 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
             mask_buf[-1] = 0;
           }
 
-          const int rc = mask_append (hashcat_ctx, mask_buf, prepend_buf);
-
-          if (rc == -1)
+          if (mask_append (hashcat_ctx, mask_buf, prepend_buf) == -1)
           {
-            fclose (mask_fp);
+            hc_fclose (&mask_fp);
 
             return -1;
           }
@@ -1597,7 +1581,7 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
 
         hcfree (line_buf);
 
-        fclose (mask_fp);
+        hc_fclose (&mask_fp);
       }
       else
       {
@@ -1617,9 +1601,7 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
 
     if (hc_path_exist (arg) == false)
     {
-      const int rc = mask_append (hashcat_ctx, arg, NULL);
-
-      if (rc == -1) return -1;
+      if (mask_append (hashcat_ctx, arg, NULL) == -1) return -1;
     }
     else
     {
@@ -1627,9 +1609,9 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
       {
         mask_ctx->mask_from_file = true;
 
-        FILE *mask_fp = fopen (arg, "r");
+        HCFILE mask_fp;
 
-        if (mask_fp == NULL)
+        if (hc_fopen (&mask_fp, arg, "r") == false)
         {
           event_log_error (hashcat_ctx, "%s: %s", arg, strerror (errno));
 
@@ -1638,9 +1620,9 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
 
         char *line_buf = (char *) hcmalloc (HCBUFSIZ_LARGE);
 
-        while (!feof (mask_fp))
+        while (!hc_feof (&mask_fp))
         {
-          const size_t line_len = fgetl (mask_fp, line_buf);
+          const size_t line_len = fgetl (&mask_fp, line_buf, HCBUFSIZ_LARGE);
 
           if (line_len == 0) continue;
 
@@ -1659,11 +1641,9 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
             mask_buf[-1] = 0;
           }
 
-          const int rc = mask_append (hashcat_ctx, mask_buf, prepend_buf);
-
-          if (rc == -1)
+          if (mask_append (hashcat_ctx, mask_buf, prepend_buf) == -1)
           {
-            fclose (mask_fp);
+            hc_fclose (&mask_fp);
 
             return -1;
           }
@@ -1671,7 +1651,7 @@ int mask_ctx_init (hashcat_ctx_t *hashcat_ctx)
 
         hcfree (line_buf);
 
-        fclose (mask_fp);
+        hc_fclose (&mask_fp);
       }
       else
       {
@@ -1700,6 +1680,9 @@ void mask_ctx_destroy (hashcat_ctx_t *hashcat_ctx)
 
   if (mask_ctx->enabled == false) return;
 
+  hcfree (mask_ctx->mp_sys);
+  hcfree (mask_ctx->mp_usr);
+
   hcfree (mask_ctx->css_buf);
 
   hcfree (mask_ctx->root_css_buf);
